from datasets import Dataset

from foggy_training.base import Task


def prepare_for_model(dataset: Dataset, task: Task):
    inputs = task.model_inputs
    target = task.model_target
    remove = [c for c in dataset.features if c not in inputs + [target]]
    dataset = dataset.remove_columns(remove)
    dataset.set_format('pt')
    return dataset



from torch import nn
from torch.nn.functional import mse_loss, cross_entropy, softmax
from transformers import AutoConfig, AutoModel

from foggy_training.base import LightningExtended, Task, TransformerTask
from foggy_training.plugins.prepare import TokenizePlugin


class SoftmaxClassifier(LightningExtended):
    def __init__(self, n_inputs: int, n_classes: int, task: Task, metrics=None):
        super().__init__(loss=cross_entropy, task=task, metrics=metrics)
        self.linear = nn.Linear(n_inputs, n_classes)

    def forward(self, x):
        x = x['x']
        x = self.linear(x)
        return softmax(x, dim=-1)


class TransformerModel(LightningExtended):
    def __init__(self, backbone: str, loss: callable, task: TransformerTask, metrics=None):
        super().__init__(loss=loss, task=task, metrics=metrics)
        config = AutoConfig.from_pretrained(backbone)
        self.backbone = backbone
        self.backbone_ = AutoModel.from_pretrained(backbone, config=config)
        self.tokenize = TokenizePlugin(self)
        self.head = task.build_head(self)

    def forward(self, x: dict):
        x = self.backbone_(input_ids=x["input_ids"], attention_mask=x["attention_mask"])
        x = self.head(x)
        return x



from typing import Iterable, List, Union, Tuple

from attr import define, field
from pytorch_lightning import LightningModule
from pytorch_lightning.callbacks.callback import Callback
from pytorch_lightning.utilities.model_summary import DeepSpeedSummary
from pytorch_lightning.utilities.model_summary import ModelSummary as Summary
from pytorch_lightning.utilities.model_summary import summarize
from pytorch_lightning.utilities.model_summary.model_summary import _format_summary_table
from torch.nn import Module, ModuleDict
from torch.nn.modules.batchnorm import _BatchNorm


@define
class FreezePlugin:
    model: LightningModule = field()

    def __call__(self, modules: Union[str, Module, Iterable[Union[Module, Iterable]]] = 'backbone',
                 include_bn: bool = True, freeze: bool = True):
        if isinstance(modules, str):
            modules = getattr(self.model, modules)
        if freeze:
            self.freeze(modules, train_bn=include_bn)
        else:
            self.unfreeze(modules)

    @staticmethod
    def flatten_modules(modules: Union[Module, Iterable[Union[Module, Iterable]]]) -> List[Module]:
        """This function is used to flatten a module or an iterable of modules into a list of its leaf modules
        (modules with no children) and parent modules that have parameters directly themselves.

        Args:
            modules: A given module or an iterable of modules

        Returns:
            List of modules
        """
        if isinstance(modules, ModuleDict):
            modules = modules.values()

        if isinstance(modules, Iterable):
            _flatten_modules = []
            for m in modules:
                _flatten_modules.extend(FreezePlugin.flatten_modules(m))

            _modules = iter(_flatten_modules)
        else:
            _modules = modules.modules()

        # Capture all leaf modules as well as parent modules that have parameters directly themselves
        return [m for m in _modules if not list(m.children()) or m._parameters]

    @staticmethod
    def unfreeze(modules: Union[Module, Iterable[Union[Module, Iterable]]]) -> None:
        """Unfreezes the parameters of the provided modules.

        Args:
            modules: A given module or an iterable of modules
        """
        modules = FreezePlugin.flatten_modules(modules)
        for module in modules:
            # recursion could yield duplicate parameters for parent modules w/ parameters so disabling it
            for param in module.parameters(recurse=False):
                param.requires_grad = True

    @staticmethod
    def freeze(modules: Union[Module, Iterable[Union[Module, Iterable]]], train_bn: bool = True) -> None:
        """Freezes the parameters of the provided modules.

        Args:
            modules: A given module or an iterable of modules
            train_bn: If True, leave the BatchNorm layers in training mode

        Returns:
            None
        """
        modules = FreezePlugin.flatten_modules(modules)
        for mod in modules:
            if isinstance(mod, _BatchNorm) and train_bn:
                FreezePlugin.unfreeze(mod)
            else:
                # recursion could yield duplicate parameters for parent modules w/ parameters so disabling it
                for param in mod.parameters(recurse=False):
                    param.requires_grad = False


@define
class SummaryPlugin:

    model: LightningModule = field()
    _max_depth: int = field(default=1)

    def __call__(self, max_depth: int = 1) -> None:
        self._max_depth: int = max_depth
        self.on_fit_start(self.model.fit.trainer, self.model)

    def on_fit_start(self, trainer, pl_module) -> None:
        if not self._max_depth:
            return None

        model_summary = self._summary(trainer, pl_module)
        summary_data = model_summary._get_summary_data()
        total_parameters = model_summary.total_parameters
        trainable_parameters = model_summary.trainable_parameters
        model_size = model_summary.model_size

        if trainer.is_global_zero:
            self.summarize(summary_data, total_parameters, trainable_parameters, model_size)

    def _summary(self, trainer, pl_module) -> Union[DeepSpeedSummary, Summary]:
        from pytorch_lightning.strategies.deepspeed import DeepSpeedStrategy

        if isinstance(trainer.strategy, DeepSpeedStrategy) and trainer.strategy.zero_stage_3:
            return DeepSpeedSummary(pl_module, max_depth=self._max_depth)
        return summarize(pl_module, max_depth=self._max_depth)

    @staticmethod
    def summarize(
            summary_data: List[Tuple[str, List[str]]],
            total_parameters: int,
            trainable_parameters: int,
            model_size: float,
    ) -> None:
        summary_table = _format_summary_table(total_parameters, trainable_parameters, model_size, *summary_data)
        print("\n" + summary_table)



from functools import singledispatchmethod

import torch
from attr import define, field
from datasets import DatasetDict
from pytorch_lightning import LightningModule
from torch.utils.data import DataLoader

from foggy_training.plugins.utils import prepare_for_model


@define
class PredictPlugin:
    """
    The FitPlugin allows you to directly train on datasets.
    """

    model: LightningModule = field()

    @singledispatchmethod
    def __call__(self, dataset, warn_prediction_exists: bool = True):
        if 'predicted' in dataset.features:
            if warn_prediction_exists:
                print(
                    'dataset has already a feature named predicted. '
                    'Did you run the evaluation previously? I skip predictions and take the existent predictions.')
            return dataset

        # disable shuffling for prediction
        config = self.model.fit.data_config.as_dict
        config = {**config, 'shuffle': False}

        # data to DataLoader
        prep = prepare_for_model(dataset, self.model.task)
        dataloader = DataLoader(prep, **config)

        predictions = self.model.fit.trainer.predict(self.model, dataloader)
        predictions = torch.cat(predictions).numpy().tolist()
        dataset = dataset.add_column('predicted', predictions)
        return dataset

    @__call__.register
    def predict_on_dataset_dict(self, dataset: DatasetDict, warn_prediction_exists: bool = True):
        for d in dataset:
            dataset[d] = self(dataset[d], warn_prediction_exists)
        return dataset



from functools import singledispatchmethod

from datasets import Dataset, DatasetDict
from transformers import AutoTokenizer


@singledispatchmethod
def analyze_length_of_text(self, dataset: Dataset):
    if 'text_length' in dataset.features:
        print('skip analysis because text_length already exists.')
        return dataset

    tokenizer = AutoTokenizer.from_pretrained(self.backbone)
    tokenized = dataset.map(
        lambda examples: tokenizer(examples[self.task.tokenize_column],
                                   padding=False, truncation=False, return_tensors="np"),
        batched=True, desc='tokenizing'
    )
    length = tokenized.map(lambda x: dict(text_length=len(x['input_ids'])))
    dataset = dataset.add_column('text_length', length['text_length'])
    return dataset


@analyze_length_of_text.register
def _(self, dataset: DatasetDict):
    for d in dataset:
        dataset[d] = self.analyze_length_of_text(dataset[d])
    return dataset





from attr import define, field
from datasets import Dataset
from pytorch_lightning import LightningModule
from transformers import AutoTokenizer


@define
class TokenizePlugin:

    model: LightningModule = field()

    def __call__(self, dataset: Dataset, padding='max_length', truncation=True, add_input_length: bool = True):
        tokenizer = AutoTokenizer.from_pretrained(self.model.backbone)
        dataset = dataset.map(
            lambda examples: tokenizer(examples[self.model.task.tokenize_column],
                                       padding=padding, truncation=truncation, return_tensors="np"),
            batched=True, desc='tokenizing'
        )

        self.model.task.model_inputs += tokenizer.model_input_names
        # list(set(...)) to remove duplicates in case the user added input names before
        self.model.task.model_inputs = list(set(self.model.task.model_inputs))

        return dataset



from functools import partial, singledispatchmethod

from attr import define, field
from datasets import Dataset, DatasetDict
from pytorch_lightning import LightningModule

from foggy_training.plugins.utils import prepare_for_model


@define
class EvaluatePlugin:

    """
    The EvaluatePlugin allows you to directly evaluate on datasets.
    """

    model: LightningModule = field()

    @singledispatchmethod
    def __call__(self, dataset: Dataset):
        if 'loss' in dataset.features:
            print('dataset has already a feature named loss. Did you run the evaluation previously?')
            return dataset

        prep = prepare_for_model(dataset, self.model.task)
        predicted = self.model.predict(prep, warn_prediction_exists=False)
        n_loss = partial(self.model.loss, reduction='none')
        y = predicted[self.model.task.model_target]
        # y = F.one_hot(predicted[self.model.task.model_target]).double()
        loss_value = n_loss(predicted['predicted'], y).numpy()

        if len(loss_value.shape) == 2:
            loss_value = loss_value.sum(axis=-1)
        if len(loss_value.shape) > 2:
            raise ValueError(f'loss has more than 2 dimensions. cannot calculate loss per instance then.')

        loss_value = loss_value.tolist()
        return dataset.add_column('loss', loss_value)

    @__call__.register
    def evaluate_on_dataset_dict(self, dataset: DatasetDict, warn_prediction_exists: bool = True):
        for d in dataset:
            dataset[d] = self(dataset[d], warn_prediction_exists)
        return dataset

    def highest_loss_instances(self, dataset: Dataset):
        if 'loss' not in dataset.features:
            dataset = self(dataset)
        dataset.set_format('pd')
        return dataset.sort('loss', reverse=True)



from typing import Union, List

import torch
from attr import field, asdict, define
from datasets import Dataset
from pytorch_lightning import Callback, LightningModule, Trainer
from torch.utils.data import DataLoader

from foggy_training.plugins.utils import prepare_for_model


@define
class DataConfig:
    batch_size: int = 32
    num_workers: int = 6
    shuffle: bool = False

    def cpu(self):
        self.num_workers = 0
        self.batch_size = 32

    def gpu(self):
        self.num_workers = 6
        self.batch_size = 32

    @property
    def as_dict(self):
        return asdict(self)  # noqa


@define
class FitPlugin:

    """
    The FitPlugin allows you to directly train on datasets.
    """

    model: LightningModule = field()
    max_epochs: int = 20
    auto_scale_batch_size: Union[str, bool] = False
    callbacks: List[Callback] = field(factory=lambda: [])
    accelerator: str = 'gpu'
    precision: Union[int, str] = 16
    logger: bool = True
    gradient_clip_val: Union[int, float] = None
    enable_progress_bar: bool = True
    overfit_batches: Union[int, float] = 0.0
    track_grad_norm: Union[int, float, str] = -1
    fast_dev_run: Union[int, bool] = False
    log_every_n_steps: int = 50
    enable_model_summary: bool = True
    num_sanity_val_steps: int = 2
    data_config: DataConfig = field(default=DataConfig())

    def __attrs_post_init__(self):
        if not torch.cuda.is_available():
            self.cpu()
            print('training config set to cpu because no gpu was detected.')

    def cpu(self):
        self.accelerator = 'cpu'
        self.precision = 32
        self.logger = False
        # self.overfit_batches = 3
        self.data_config.cpu()

    def gpu(self):
        self.accelerator = 'gpu'
        self.precision = 16
        self.logger = True
        self.overfit_batches = 0.0
        self.data_config.gpu()

    @property
    def as_dict(self):
        exclude = ['model', 'data_config']
        return asdict(self, filter=lambda attr, value: attr.name not in exclude)  # noqa

    @property
    def trainer(self) -> Trainer:
        return Trainer(**self.as_dict)

    def __call__(self, train_dataset: Dataset, epochs: int = 20, val_dataset: Union[Dataset, float, int] = None):
        train_dataset = prepare_for_model(train_dataset, self.model.task)

        if val_dataset is not None:
            if isinstance(val_dataset, float) or isinstance(val_dataset, int):
                val_dataset = train_dataset.train_test_split(test_size=val_dataset, seed=31)

        self.data_config.shuffle = True
        train_dataset = prepare_for_model(train_dataset, self.model.task)
        train_dataloader = DataLoader(train_dataset, **self.data_config.as_dict)

        if val_dataset is not None:
            val_dataset = prepare_for_model(val_dataset, self.model.task)
            cfg = self.data_config.as_dict
            cfg['shuffle'] = False
            val_dataset = DataLoader(val_dataset, **cfg)

        self.max_epochs = epochs

        self.trainer.fit(self.model, train_dataloaders=train_dataloader, val_dataloaders=val_dataset)



import pytorch_lightning as pl
from torch import optim
from torchmetrics import MetricCollection

from foggy_training.base import Task
from foggy_training.plugins.evaluate import EvaluatePlugin
from foggy_training.plugins.fit import FitPlugin
from foggy_training.plugins.freeze import FreezePlugin, SummaryPlugin
from foggy_training.plugins.prediction import PredictPlugin


class LightningExtended(pl.LightningModule):

    """

    The Lightning Extended Modules provides some defaults for training and validation sets. You may do the following:

    ```python
    from foggy_training import models
    x = [[1., 1.], [2., 2.], [3., 3.]]
    y = [1., 2., 3.]
    data = {'x': x, 'y': y}
    model = models.LinearRegression(n_inputs=2)
    model.fit(data, epochs=20)
    ```

    """
    def __init__(self, loss: callable, task: Task, metrics=None):
        super(LightningExtended, self).__init__()
        self.task: Task = task
        self.fit = FitPlugin(self)
        self.predict = PredictPlugin(self)
        self.evaluate = EvaluatePlugin(self)
        self.freeze = FreezePlugin(self)
        self.summary = SummaryPlugin(self)
        self.loss = loss
        self.lr = 1e-3
        metrics = [] if metrics is None else metrics
        assert isinstance(metrics, list), 'metrics must be list of metrics'
        metrics = MetricCollection(metrics)
        self.train_metrics = metrics.clone(prefix='train_')
        self.val_metrics = metrics.clone(prefix='val_')

    def shared_step(self, batch):
        x = {k: batch[k] for k in self.task.model_inputs}
        y = batch.get(self.task.model_target)
        pred = self(x)
        return pred, y

    def training_step(self, batch, batch_idx):
        pred, y = self.shared_step(batch)
        loss = self.loss(pred, y)
        logs = dict(loss=loss)
        metrics = self.train_metrics(pred, y)
        logs = {**logs, **metrics}
        self.log_dict(logs, prog_bar=True, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        pred, y = self.shared_step(batch)
        loss = self.loss(pred, y)
        logs = dict(val_loss=loss)
        metrics = self.val_metrics(pred, y)
        logs = {**logs, **metrics}
        self.log_dict(logs, prog_bar=True, on_epoch=True)

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        return self.shared_step(batch)[0]

    def configure_optimizers(self):
        return optim.AdamW(filter(lambda p: p.requires_grad, self.parameters()), lr=self.lr)



from typing import List, Union

from attr import define, field
from torch import nn


@define
class Task:
    model_inputs: Union[List[str], str] = field(converter=lambda v: v if isinstance(v, list) else [v])
    model_target: str = field()


@define
class MultiOutputRegression(Task):
    pass


@define
class Classification(Task):
    pass


@define
class TransformerTask(Task):
    tokenize_column: str = field()

    def build_head(self, model):
        input_size = model.backbone_.config.hidden_size
        hidden_size = 150
        num_layers = 2
        num_classes = 6

        class LambdaLayer(nn.Module):
            def __init__(self, lambd):
                super(LambdaLayer, self).__init__()
                self.lambd = lambd

            def forward(self, x):
                return self.lambd(x)

        return nn.Sequential(
            LambdaLayer(lambda x: x.last_hidden_state),
            nn.LSTM(input_size=input_size, hidden_size=hidden_size, num_layers=num_layers, batch_first=True),
            LambdaLayer(lambda x: x[0]),
            nn.Linear(hidden_size, hidden_size // 2),
            nn.Flatten(),
            nn.LazyLinear(num_classes)
        )

